import pandas as pd
import numpy as np
import json
from .fund_info_filter import EHR_PROBLEM_EKYWORDS

def fund_tag(base_fund_id:str,  #基金母基金代码 非母基金是0
             desc_name:str,     #基金中文简称
             is_open:int,       #是否是开放式基金  1表示开放 0 表示定期开放
             currency:str,      #交易货币类型
             wind_class_2:str,  #wind二级分类
             benchmark_s_raw:str,    #basic.fund_benchmark.benchmark_s_raw
             fund_id:str,       #fund_id
):
    
    '''
    删除基金逻辑
    '''

    # 1.只选择非分级基金
    ## wind 基金信息数据里 基金母基金代码 如果如果base_fund_id 为0表示 非分级基金， 
    ##                               如果base_fund_id 和 基金代码一致，表示是分级基金母基金
    ##                               如果不一致，可以通过desc_name区分 A类分级或者B类分级
    if base_fund_id != 0:
        return None

    # 2.只选择非ETF基金,可以选择ETF联接
    _desc_name = desc_name.replace('ETF联接','')
    if 'ETF' in _desc_name:
        return None

    # 3.只选择开放式基金 不选择定期开放式基金
    if is_open != 1:
        return None

    # 4.只选择人民币交易的基金
    if currency != 'CNY':
        return None

    # 5.去掉基金份额 E类 H类 R类
    ## E类 指定代销机构对基金
    ## H类 基金场内上市份额
    ## R类 对特定投资群体进行发售，特定群体指基本养老保险基金，企业年金计划筹集的资金等

    for type_word, problem_words in EHR_PROBLEM_EKYWORDS.items():
        for problem_i in problem_words:
            _desc_name = desc_name.replace(problem_i, '')
        if type_word in _desc_name:
            return None

    # 6.国债里不包含短债 短融 双债 短期纯债
    debt_black_list = ['超短债','短融','双债']
    for i in debt_black_list:
        if i in desc_name:
            return None
    if wind_class_2 == '短期纯债型基金':
        return None

    '''
    分类逻辑 分类计数在不考虑删除逻辑统计
    '''
    # 1. 沪深300  141 只 hs300 000300.SH
    ## 第一基准是hs300 即权重大于50%
    ## wind 二级投资分类 在 【'增强指数型基金','被动指数型基金'】 不包含 【‘普通股票型基金’,'偏股混合型基金'】
    ## 【'增强指数型基金','被动指数型基金'】 里hs300权重最低 0.9
    hs300_w = json.loads(benchmark_s_raw).get('hs300',0)
    if hs300_w > 0.5 and wind_class_2 in ['增强指数型基金','被动指数型基金']:
        return 'hs300'

    # 2. 中证500 104只 csi500 000905.SH
    ## 第一基准是csi500 即权重大于50%
    ## wind 二级投资分类 在 【'增强指数型基金','被动指数型基金'】 不包含 【‘普通股票型基金’,'偏股混合型基金'】
    ## 【'增强指数型基金','被动指数型基金'】 里hs300权重最低 0.95
    csi500_w = json.loads(benchmark_s_raw).get('csi500',0)
    if csi500_w > 0.5 and wind_class_2 in ['增强指数型基金','被动指数型基金']:
        return 'csi500'

    # 3. 创业板 57 只 gem 
    ## 创业板基金少，考虑所有创业板相关基准
    ## 399295.SZ	cnext_lowv_bc   创业蓝筹     
    ## 399606.SZ	cnext_mome_gr   创业板R      
    ## 399296.SZ	cnext_r         创成长       
    ## 399102.SZ	gei             创业板综指   
    ## 399673.SZ	gem             创业板50    
    ## 399006.SZ	gem_50          创业板指    
    ## 只选 ['增强指数型基金','被动指数型基金']

    gem_list = ['cnext_lowv_bc','cnext_mome_gr','cnext_r','gei','gem','gem_50']
    con1 = False
    for gem_id in gem_list:
        w = json.loads(benchmark_s_raw).get(gem_id,0)
        if w > 0.5:
            con1 = True
            continue
    con2 = wind_class_2 in ['增强指数型基金','被动指数型基金']
    if con1 and con2:
        return 'gem'

    # 4. 标普500 8只 sp500rmb
    ## 第一基准是标普
    sp500rmb_w = json.loads(benchmark_s_raw).get('sp500rmb',0)
    if sp500rmb_w > 0.5 and wind_class_2 in ['增强指数型基金','被动指数型基金']:
        return 'sp500rmb'

    # 5. 国债 2129 只 national_debt
    ## 选取第一基准不是行业债 金融债的 
    ## cba_10ytfp_all              中债10年期国债全价(总值)指数
    ## cba_10y_future              中债10年期国债期货期限匹配金融债全价(总值)指数
    ## cba_130                     中债1-30年利率债财富(1-3年)指数
    ## cba_139qi                   中债-1-3年久期央企20债券财富(总值)指数
    ## cba_5y_future               中债5年期国债期货期限匹配金融债全价(总值)指数
    ## cba_a_fp_1_3y               中债总全价(1-3年)指数
    ## cba_a_fp_all                中债总全价(总值)指数
    ## cba_a_tr_1_3y               中债总财富(1-3年)指数
    ## cba_a_tr_all                中债总财富(总值)指数
    ## cba_crefp_all               中债信用债总全价(总值)指数
    ## cba_cretr_1y                中债信用债总财富(1年以下)指数
    ## cba_cretr_1_3y              中债信用债总财富(1-3年)指数
    ## cba_cretr_3_5y              中债信用债总财富(3-5年)指数
    ## cba_cretr_all               中债信用债总财富(总值)指数
    ## cba_c_fp_1y                 中债综合全价(1年以下)指数
    ## cba_c_fp_1_3y               中债综合全价(3-5年)指数
    ## cba_c_fp_all                中债综合全价(总值)指数
    ## cba_c_tr_1y                 中债综合财富(1年以下)指数
    ## cba_c_tr_1_3y               中债综合财富(1-3年)指数
    ## cba_c_tr_all                中债综合财富(总值)指数
    ## cba_e_fp_all                中债企业债总全价(总值)指数
    ## cba_f_fp_1_3y               中债金融债券总全价(总值)指数
    ## cba_f_tr_all                中债金融债券总财富(总值)指数
    ## cba_hg_fp_1_3y              中债高信用等级债券财富(1-3年)指数
    ## cba_hg_fp_all               中债高信用等级债券全价(总值)指数
    ## cba_hg_tr_all               中债高信用等级债券财富(总值)指数
    ## cba_nc_fp_1y                中债新综合全价(1年以下)指数
    ## cba_nc_fp_1_3y              中债新综合全价(1-3年)指数
    ## cba_nc_fp_all               中债新综合全价(总值)指数
    ## cba_nc_tr_1_3y              中债新综合财富(1-3年)指数
    ## cba_nc_tr_all               中债新综合财富(总值)指数
    ## cba_sm_1_3y                 中债中短期债券财富(1-3年)指数
    ## cba_t_tr_1_3y               中债国债总财富(1-3年)指数
    ## cba_t_tr_3_5y               中债国债总财富(3-5年)指数
    ## cba_t_tr_5_10y              中债国债总财富(7-10年)指数
    ## cba_t_tr_all                中债国债总财富(总值)指数
    ## cni_tpbb                    国证利率
    ## cni_tpbb5                   国证利率1-5
    ## csi_10ytb                   中证10年国债
    ## csi_5ydcdb                  5年久期国开债
    ## csi_cib_mh_b                中证兴业中高信用债
    ## csi_compb                   中证综合债
    ## csi_hcityz                  沪城投债
    ## csi_medcomz                 中期企债
    ## csi_meddnb                  中期国债
    ## csi_nationb                 中证国债
    ## csi_patb                    中证平安5-10年国债
    ## csi_pbb                     政金债
    ## csi_pbb10                   政金债8-10
    ## csi_pbb3                    政金债1-3
    ## csi_pbb5                    政金债3-5
    ## csi_pbb8                    政金债5-8
    ## csi_wholeb                  中证全债
    ## national_debt_index         国债指数
    ## see_3_5y_creb               沪质中高债3-5
    ## sse_nationb5                5年国债
    ## sz_national_10y             10年期国债
    ## tmd_1y                      一年期定期存款利率
    ## tmd_2y                      两年期定期存款利率
    ## tmd_3m                      三个月定期存款利率
    ## tmd_3y                      三年期定期存款利率
    ## tmd_6m                      六个月定期存款利率
    ## tmd_7d                      七天通知存款利率

    ## 选择wind 二级分类 ['被动指数型债券基金','中长期纯债型基金']

    national_index_list = ['cba_10ytfp_all',
                            'cba_10y_future',
                            'cba_130',
                            'cba_139qi',
                            'cba_5y_future',
                            'cba_a_fp_1_3y',
                            'cba_a_fp_all',
                            'cba_a_tr_1_3y',
                            'cba_a_tr_all',
                            'cba_crefp_all',
                            'cba_cretr_1y',
                            'cba_cretr_1_3y',
                            'cba_cretr_3_5y',
                            'cba_cretr_all',
                            'cba_c_fp_1y',
                            'cba_c_fp_1_3y',
                            'cba_c_fp_all',
                            'cba_c_tr_1y',
                            'cba_c_tr_1_3y',
                            'cba_c_tr_all',
                            'cba_e_fp_all',
                            'cba_f_fp_1_3y',
                            'cba_f_tr_all',
                            'cba_hg_fp_1_3y',
                            'cba_hg_fp_all',
                            'cba_hg_tr_all',
                            'cba_nc_fp_1y',
                            'cba_nc_fp_1_3y',
                            'cba_nc_fp_all',
                            'cba_nc_tr_1_3y',
                            'cba_nc_tr_all',
                            'cba_sm_1_3y',
                            'cba_t_tr_1_3y',
                            'cba_t_tr_3_5y',
                            'cba_t_tr_5_10y',
                            'cba_t_tr_all',
                            'cni_tpbb',
                            'cni_tpbb5',
                            'csi_10ytb',
                            'csi_5ydcdb',
                            'csi_cib_mh_b',
                            'csi_compb',
                            'csi_hcityz',
                            'csi_medcomz',
                            'csi_meddnb',
                            'csi_nationb',
                            'csi_patb',
                            'csi_pbb',
                            'csi_pbb10',
                            'csi_pbb3',
                            'csi_pbb5',
                            'csi_pbb8',
                            'csi_wholeb',
                            'national_debt_index',
                            'see_3_5y_creb',
                            'sse_nationb5',
                            'sz_national_10y',
                            'tmd_1y',
                            'tmd_2y',
                            'tmd_3m',
                            'tmd_3y',
                            'tmd_6m',
                            'tmd_7d']   
    national_bench = json.loads(benchmark_s_raw)
    index_list = []
    for index_id, w in national_bench.items():
        if w >= 0.5:
            index_list.append(index_id)
    con1 = False
    for i in national_index_list:
        if i in index_list:
            con1 = True
            break
    con2 = wind_class_2 in ['被动指数型债券基金','中长期纯债型基金']
    if con1 and con2:
        return 'national_debt'
    
    # 6. 货基
    ## 在wind 一级分类下选择 货币市场型基金, 然后在2020年5月截面上，选择规模最大的前50只
    ## 000198!0     天弘余额宝
    ## 000330!0     汇添富现金宝
    ## 000343!0     华夏财富宝A
    ## 000359!0     易方达易理财A
    ## 000379!0     平安日增利
    ## 000380!0     景顺长城景益货币A
    ## 000397!0     汇添富全额宝
    ## 000539!0     中银活期宝
    ## 000559!0     诺安天天宝A
    ## 000569!0     鹏华增值宝
    ## 000575!0     兴全添利宝
    ## 000581!0     嘉实活钱包A
    ## 000607!0     招商招钱宝B
    ## 000621!0     易方达现金增利B
    ## 000638!0     富国富钱包
    ## 000662!0     银华活钱宝F
    ## 000693!0     建信现金添利A
    ## 000719!0     南方现金通E
    ## 000759!0     平安财富宝
    ## 000848!0     工银瑞信添益快线
    ## 000907!0     农银汇理红利A
    ## 000917!0     嘉实快线A
    ## 001134!0     广发天天利E
    ## 001211!0     中欧滚钱宝A
    ## 001529!0     天弘云商宝
    ## 001666!0     鹏华添利宝
    ## 001821!0     兴全天添益B
    ## 002758!0     建信现金增利货币
    ## 003164!0     建信现金添利B
    ## 003281!0     广发活期宝B
    ## 003474!0     南方天天利B
    ## 003515!0     国泰利是宝
    ## 003753!0     工银瑞信如意B
    ## 004137!0     博时合惠B
    ## 004199!0     华富天益B
    ## 004217!0     兴业安润B
    ## 004417!0     兴全货币B
    ## 004501!0     嘉实现金添利
    ## 004545!0     永赢天天利
    ## 004771!0     海富通添益货币B
    ## 004776!0     鹏华金元宝
    ## 040038!0     华安日日鑫A
    ## 050003!0     博时现金收益A
    ## 180008!0     银华货币A
    ## 200003!0     长城货币A
    ## 37001B!0     上投摩根货币B
    ## 482002!0     工银瑞信货币
    ## 511880!0     银华交易货币A
    ## 511990!0     华宝现金添益A
    ## 519510!0     浦银安盛货币B
    
    mmf_list = ['000198!0',
                '000330!0',
                '000343!0',
                '000359!0',
                '000379!0',
                '000380!0',
                '000397!0',
                '000539!0',
                '000559!0',
                '000569!0',
                '000575!0',
                '000581!0',
                '000607!0',
                '000621!0',
                '000638!0',
                '000662!0',
                '000693!0',
                '000719!0',
                '000759!0',
                '000848!0',
                '000907!0',
                '000917!0',
                '001134!0',
                '001211!0',
                '001529!0',
                '001666!0',
                '001821!0',
                '002758!0',
                '003164!0',
                '003281!0',
                '003474!0',
                '003515!0',
                '003753!0',
                '004137!0',
                '004199!0',
                '004217!0',
                '004417!0',
                '004501!0',
                '004545!0',
                '004771!0',
                '004776!0',
                '040038!0',
                '050003!0',
                '180008!0',
                '200003!0',
                '37001B!0',
                '482002!0',
                '511880!0',
                '511990!0',
                '519510!0']
                    
    if fund_id in mmf_list:
        return 'mmf'

    # 7. gold 
    if '黄金' in desc_name:
        return 'gold' 
    return None



from typing import List, Dict
import pandas as pd
import numpy as np
import datetime
import traceback
import json
import math
import statsmodels.api as sm
from scipy.optimize import Bounds, minimize
from statsmodels.tsa.ar_model import AutoReg
from ...api.basic import BasicDataApi


class FundIndicatorProcessorGroup:

    _TRADING_DAYS_PER_YEAR = 242
    _NATURAL_DAYS_PER_YEAR = 365
    _RISK_FEE_RATE = 0.025
    _RISK_FEE_RATE_PER_DAY = _RISK_FEE_RATE / _TRADING_DAYS_PER_YEAR

    def __init__(self, data_helper):
        self._data_helper = data_helper
        self._basic_api = BasicDataApi()
        self._rank_dic: Dict[pd.DataFrame] = {}
        self._res: List[pd.Series] = []

    def init(self, start_date: str, end_date: str):
        # self.log = []
        # 向前取2年
        start_date: str = (pd.to_datetime(start_date, infer_datetime_format=True) - datetime.timedelta(days=self._NATURAL_DAYS_PER_YEAR*2+1)).strftime('%Y%m%d')
        print(f'start date for data: {start_date}')
        # 获取区间内交易日列表
        all_trading_days: pd.Series = self._basic_api.get_trading_day_list().drop(columns='_update_time').set_index('datetime')
        self._trading_days: pd.Series = all_trading_days.loc[pd.to_datetime(start_date, infer_datetime_format=True).date():pd.to_datetime(end_date, infer_datetime_format=True).date()]
        # 使用实际trading day的最后一天作为end_date
        self.end_date = self._trading_days.index.array[-1]

        # 基金净值数据、指数价格数据的index对齐到交易日列表
        fund_nav: pd.DataFrame = self._basic_api.get_fund_nav_with_date(start_date, end_date)  #, fund_list = f_l)
        self._fund_nav: pd.DataFrame = fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').reindex(self._trading_days.index).fillna(method='ffill')
        # 选取成立满半年的 而且近半年波动率不为0，并排除掉异常净值基金
        res = []
        for fund_id in self._fund_nav:
            _df = self._fund_nav[fund_id].dropna()
            if _df.shape[0] > 120:
                if _df.iloc[-120:].var() > 1e-6:
                    if _df.pct_change(1).abs().max() < 0.2:
                        res.append(fund_id)
        self._fund_nav = self._fund_nav[res].copy()

        index_price: pd.DataFrame = self._basic_api.get_index_price_dt(start_date, end_date).drop(columns='_update_time')
        # 有的index最近没有数据，如活期存款利率/各种定期存款利率等，需要先reindex到全部的交易日列表上ffill
        self._index_price: pd.DataFrame = index_price.pivot_table(index='datetime', columns='index_id', values='close').reindex(self._trading_days.index)  #.fillna(method='ffill')
        _bank_rate_df: pd.DataFrame = self._basic_api.get_index_price(index_list=['ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d']).drop(columns='_update_time')
        _bank_rate_df = _bank_rate_df.pivot_table(index='datetime', columns='index_id', values='close').reindex(all_trading_days.index).fillna(method='ffill').reindex(self._trading_days.index)
        self._index_price: pd.DataFrame = pd.concat([self._index_price, _bank_rate_df], axis=1)
        # 再reindex到参与计算的交易日列表上
        # self._index_price: pd.DataFrame = index_price.reindex(self._trading_days.index)
        pd.testing.assert_index_equal(self._fund_nav.index, self._index_price.index)
        try:
            # 这个指数有一天的点数是0，特殊处理一下
            self._index_price['spi_spa'] = self._index_price['spi_spa'].where(self._index_price.spi_spa != 0).fillna(method='ffill')
        except KeyError:
            pass

        # 对数净值取差分并且去掉第一个空值，得到基金对数收益率数列
        self._fund_ret: pd.DataFrame = np.log(self._fund_nav).diff().iloc[1:, :]

        self._fund_size: pd.DataFrame = self._basic_api.get_fund_size_range(start_date, end_date)
        self._fund_size = self._fund_size.pivot_table(index='datetime', columns='fund_id', values='size').reindex(self._fund_ret.index).fillna(method='ffill')
        pd.testing.assert_index_equal(self._fund_ret.index, self._fund_size.index)

        # 计算benchmark return，且index对齐到fund_ret
        benchmark_ret: pd.DataFrame = self._get_benchmark_return()
        self._benchmark_ret: pd.DataFrame = benchmark_ret.reindex(self._fund_ret.index)
        pd.testing.assert_index_equal(self._fund_ret.index, self._benchmark_ret.index)

        # 获取待计算的基金列表
        fund_list: pd.DataFrame = self._basic_api.get_fund_info().drop(columns='_update_time')
        self._fund_list: pd.DataFrame = fund_list[fund_list.structure_type <= 1]
        # 获取wind一级分类
        self._wind_class_1: np.ndarray = fund_list.wind_class_1.unique()
        # 按wind二级分类计算分类呢基金收益率排名
        self._get_continue_stats_tool()

    def df_resample(self, df):
        df.index = pd.to_datetime(df.index)
        df = df.resample('1W').sum()
        df.index = df.index.date
        return df

    def stutzer_index(self, ex_return: np.ndarray):
        information_statistic = lambda theta: np.log(np.mean(np.exp(theta[0] * ex_return)))
        theta0 = [-1.]
        bounds = Bounds((-np.inf), (0))
        result = minimize(information_statistic, theta0, method='SLSQP', bounds=bounds, tol=1e-16)
        if result.success:
            information_statistic = -result.fun
            if information_statistic <= 0:
                return np.nan
            stutzer_index = np.abs(np.mean(ex_return)) / np.mean(ex_return) * np.sqrt(2 * information_statistic)
            # result = {'information_statistic': information_statistic,
            #         'stutzer_index': stutzer_index,
            #         'theta': result.x[0]}
            return stutzer_index
        else:
            # result = {'information_statistic': information_statistic,
            #         'stutzer_index': np.nan,
            #         'theta': result.x[0]}
            # print('未找到Information Statistic最大值')
            return np.nan

    def _lambda_cl(self, total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X = np.array([total[:, 0], total[:, 0]]).T
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X[:, 0][X[:, 0] < 0] = 0
        est2 = sm.OLS(total[:, 1], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                'alpha_t':est2.tvalues[-1],
                'alpha_p':est2.pvalues[-1],
                # TODO beta 2 - beta 1 , p values and t values calculation is not correct
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }

    def _lambda_tm(self, total: np.ndarray):
        if not total.size:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X = np.array([total[:, 0], total[:, 0] * total[:, 0]]).T
        est2 = sm.OLS(total[:, 1], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[1],
                'alpha':est2.params[-1],
                'alpha_t':est2.tvalues[-1],
                'alpha_p':est2.pvalues[-1],
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }

    def _lambda_hm(self, total: np.ndarray):
        if not total.size:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X = np.array([total[:, 0], total[:, 0]]).T
        X[:, 1][X[:, 1] < 0] = 0
        est2 = sm.OLS(total[:, 1], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[1],
                'alpha':est2.params[-1],
                'alpha_t':est2.tvalues[-1],
                'alpha_p':est2.pvalues[-1],
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }

    def _lambda_alpha_beta(self, df_i):
        if np.count_nonzero(np.count_nonzero(df_i, axis=0)) < df_i.shape[1]:
            return np.nan, 0
        ploy_res = np.polyfit(y=df_i[:, 0], x=df_i[:, 1], deg=1)
        beta = ploy_res[0]
        alpha = ploy_res[1] * self._TRADING_DAYS_PER_YEAR
        return alpha, beta

    def _get_benchmark_return(self) -> pd.DataFrame:
        benchmark_list: Dict[str, float] = {}
        fund_benchmark_df: pd.DataFrame = self._basic_api.get_fund_benchmark().drop(columns='_update_time')
        # 遍历每一只基金的benchmark进行处理
        for row in fund_benchmark_df.itertuples(index=False):
            values: List[pd.Series] = []
            cons: float = 0
            # 空的benchmark表明我们没有对应的指数或无法解析公式
            if row.benchmark_s:
                benchmark: Dict[str, float] = json.loads(row.benchmark_s)
                benchmark_raw: Dict[str, float] = eval(row.benchmark)
                for (index, weight), index_raw in zip(benchmark.items(), benchmark_raw.keys()):
                    if index == '1':
                        # 表示benchmark中该项为常数
                        cons += weight
                    elif index in ('ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'):
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            # self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                ra: pd.Series = self._index_price.loc[:, index].copy()
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'[benchmark_return] Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                # self.log.append((row.fund_id, index))
                                break
                            else:
                                values.append(ra.iloc[1:] * 0.01 * weight / self._TRADING_DAYS_PER_YEAR)
                    else:
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            # self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                ra: pd.Series = self._index_price.loc[:, index].copy()
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                # self.log.append((row.fund_id, index))
                                break
                            else:
                                ra = np.log(ra).diff().iloc[1:]
                                values.append(ra * weight)
                else:
                    if values or cons:
                        the_sum: float = sum(values)
                        if cons:
                            the_sum += np.log(math.pow(1 + cons, 1 / self._TRADING_DAYS_PER_YEAR))
                        benchmark_list[row.fund_id] = the_sum

        return pd.DataFrame.from_dict(benchmark_list)

    def _lambda_make_rank_dic(self, x: pd.DataFrame):
        resample_ret: pd.DataFrame = self.df_resample(self._fund_ret.loc[:, self._fund_ret.columns.intersection(x.fund_id.array)])
        self._rank_dic[x.name] = resample_ret.rank(pct=True, axis=1).rename_axis(columns='fund_id')

    def _get_continue_stats_tool(self):
        self._fund_list.groupby(by='wind_class_2').apply(self._lambda_make_rank_dic)

    def get_scale(self, end_date, fund_id):
        try:
            scale = self._fund_size.loc[self.end_date, fund_id]
        except KeyError:
            scale = np.nan
        return scale

    def _calculate_item(self, fund_id: str, wind_2_class: str):
        # 取最后一年的有效净值
        _sr: pd.Series = self._fund_nav[fund_id]
        sr: pd.Series = _sr[_sr.notna()].tail(self._TRADING_DAYS_PER_YEAR)
        temp_nav = _sr.to_numpy()
        # 生成一个18*n的矩阵
        temp_nav = np.array([temp_nav[i*20:self._TRADING_DAYS_PER_YEAR//2+(i+1)*20] for i in range(18)])
        ret: np.ndarray = temp_nav[:, 1:] / temp_nav[:, :-1]
        # 过滤掉全是nan或只有1个不是nan的行
        filter1 = (ret.shape[1] - np.count_nonzero(np.isnan(ret), axis=1) > 1)
        ret, temp_nav = ret[filter1], temp_nav[filter1]
        annual_vol: np.ndarray = np.nanstd(ret, axis=1, ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        # 一维数组，所以这里直接用nonzero即可
        filter1 = annual_vol.nonzero()
        annual_vol, temp_nav = annual_vol[filter1], temp_nav[filter1]
        annual_ret = np.exp(np.log(temp_nav[:, -1] / temp_nav[:, 0]) / temp_nav.shape[1] * self._TRADING_DAYS_PER_YEAR) - 1
        _res = (annual_ret - self._RISK_FEE_RATE) / annual_vol
        _res = _res[~np.isnan(_res)]
        if _res.size < 6:
            continue_regress_v = np.nan
            continue_regress_t = np.nan
        else:
            mod = AutoReg(_res, 1)
            res = mod.fit()
            continue_regress_v = res.params[0]
            continue_regress_t = res.tvalues[0]

        _rank_df = self._rank_dic[wind_2_class]
        res = []
        last = ''
        for idx, i in enumerate(_rank_df[fund_id]):
            now = 'W' if i > 0.5 else 'L'
            if idx != 0:
                res.append(last+now)
            last = now
        ww = max(res.count('WW'), 1)
        ll = max(res.count('LL'), 1)
        wl = max(res.count('WL'), 1)
        lw = max(res.count('LW'), 1)
        continue_stats_v = ww * ll / wl / lw

        # 取最后一年的有效收益率+业绩比较基准收益率
        df_i: np.ndarray = self._fund_ret[[fund_id]].join(self._benchmark_ret[fund_id], rsuffix='_b').to_numpy()
        df_i = df_i[(~np.isnan(df_i)).all(axis=1)]
        df_i = df_i[-self._TRADING_DAYS_PER_YEAR:]

        # alpha, beta
        alpha, beta = self._lambda_alpha_beta(df_i)
        # mdd
        mdd_part = sr / sr.cummax()
        mdd = 1 - mdd_part.min()
        # sharpe ratio
        annual_vol = (sr / sr.shift(1)).std(ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        annual_ret = np.exp(np.log(sr.iloc[-1] / sr.fillna(method='bfill').iloc[0]) / sr.shape[0] * self._TRADING_DAYS_PER_YEAR) - 1
        sharpe = (annual_ret - self._RISK_FEE_RATE) / annual_vol
        # treynor ratio
        if beta != 0:
            treynor = (np.nanmean(df_i[:, 0]) - self._RISK_FEE_RATE_PER_DAY) / beta
        else:
            treynor = np.nan
        # calma ratio
        calma_ratio = annual_ret / mdd if mdd != 0 else np.nan
        # ex return
        _ex_return = df_i[:, 0] - df_i[:, 1]
        # track error
        track_err = np.nanstd(_ex_return, ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        # information error
        info_ratio = alpha / track_err
        # scale
        scale = self.get_scale(self.end_date, fund_id)

        stock_tm_res = self._lambda_tm(df_i)
        stock_tm_alpha = stock_tm_res['alpha']
        stock_tm_alpha_t = stock_tm_res['alpha_t']
        stock_tm_alpha_p = stock_tm_res['alpha_p']
        stock_tm_beta = stock_tm_res['beta']
        stock_tm_beta_t = stock_tm_res['beta_t']
        stock_tm_beta_p = stock_tm_res['beta_p']

        stock_hm_res = self._lambda_hm(df_i)
        stock_hm_alpha = stock_hm_res['alpha']
        stock_hm_alpha_t = stock_hm_res['alpha_t']
        stock_hm_alpha_p = stock_hm_res['alpha_p']
        stock_hm_beta = stock_hm_res['beta']
        stock_hm_beta_t = stock_hm_res['beta_t']
        stock_hm_beta_p = stock_hm_res['beta_p']

        stock_cl_res = self._lambda_cl(df_i)
        stock_cl_alpha = stock_cl_res['alpha']
        stock_cl_alpha_t = stock_cl_res['alpha_t']
        stock_cl_alpha_p = stock_cl_res['alpha_p']
        stock_cl_beta = stock_cl_res['beta']
        stock_cl_beta_t = stock_cl_res['beta_t']
        stock_cl_beta_p = stock_cl_res['beta_p']

        # stutzer
        stutzer = self.stutzer_index(_ex_return)
        # stutzer = self.stutzer_index(np.array(self._fund_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id] - self._benchmark_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id]))
        return pd.Series({
            'fund_id':fund_id,
            'datetime':self.end_date,
            'alpha':alpha,
            'beta':beta,
            'annual_vol':annual_vol,
            'annual_ret':annual_ret,
            'track_err':track_err,
            'mdd': mdd,
            'continue_regress_v':continue_regress_v,
            'continue_regress_t':continue_regress_t,
            'continue_stats_v':continue_stats_v,

            'stock_cl_alpha':stock_cl_alpha,
            'stock_cl_alpha_t':stock_cl_alpha_t,
            'stock_cl_alpha_p':stock_cl_alpha_p,
            'stock_cl_beta':stock_cl_beta,
            'stock_cl_beta_t':stock_cl_beta_t,
            'stock_cl_beta_p':stock_cl_beta_p,

            'stock_tm_alpha':stock_tm_alpha,
            'stock_tm_alpha_t':stock_tm_alpha_t,
            'stock_tm_alpha_p':stock_tm_alpha_p,
            'stock_tm_beta':stock_tm_beta,
            'stock_tm_beta_t':stock_tm_beta_t,
            'stock_tm_beta_p':stock_tm_beta_p,

            'stock_hm_alpha':stock_hm_alpha,
            'stock_hm_alpha_t':stock_hm_alpha_t,
            'stock_hm_alpha_p':stock_hm_alpha_p,
            'stock_hm_beta':stock_hm_beta,
            'stock_hm_beta_t':stock_hm_beta_t,
            'stock_hm_beta_p':stock_hm_beta_p,
            'info_ratio':info_ratio,
            'treynor':treynor,
            'sharpe':sharpe,
            'calma_ratio':calma_ratio,
            'stutzer':stutzer,
            'scale':scale,
        })

    def calculate(self):
        fund_list = self._fund_list[['fund_id', 'wind_class_2']]
        # 对fund_nav, benchmark_ret, fund_list三者交集的每个fund_id进行计算
        self._result = fund_list[fund_list.fund_id.isin(self._fund_nav.columns.intersection(self._benchmark_ret.columns).to_list())].apply(lambda x: self._calculate_item(x.fund_id, x.wind_class_2), axis=1)

    def process(self, start_date: str, end_date: str, end_date_dt: datetime.date) -> List[str]:
        print(f'indicator group update on the last day of week {end_date_dt}')
        failed_tasks: List[str] = []
        try:
            self.init(start_date, end_date)
            self.calculate()
            self._data_helper._upload_derived(self._result, FundIndicatorGroup.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_indicator_group')
        return failed_tasks
